"""Tests for the ConfigFileFinder."""
import configparser
import os
from unittest import mock

import pytest

from flake8.options import config

CLI_SPECIFIED_FILEPATH = "tests/fixtures/config_files/cli-specified.ini"
BROKEN_CONFIG_PATH = "tests/fixtures/config_files/broken.ini"


def test_cli_config():
    """Verify opening and reading the file specified via the cli."""
    cli_filepath = CLI_SPECIFIED_FILEPATH
    finder = config.ConfigFileFinder("flake8")

    parsed_config = finder.cli_config(cli_filepath)
    assert parsed_config.has_section("flake8")


@pytest.mark.parametrize(
    "cwd,expected",
    [
        # Root directory of project
        (
            os.path.abspath("."),
            [os.path.abspath("setup.cfg"), os.path.abspath("tox.ini")],
        ),
        # Subdirectory of project directory
        (
            os.path.abspath("src"),
            [os.path.abspath("setup.cfg"), os.path.abspath("tox.ini")],
        ),
        # Outside of project directory
        (os.path.abspath("/"), []),
    ],
)
def test_generate_possible_local_files(cwd, expected):
    """Verify generation of all possible config paths."""
    finder = config.ConfigFileFinder("flake8")

    with mock.patch.object(os, "getcwd", return_value=cwd):
        config_files = list(finder.generate_possible_local_files())

    assert config_files == expected


@pytest.mark.parametrize(
    "extra_config_files,expected",
    [
        # Extra config files specified
        (
            [CLI_SPECIFIED_FILEPATH],
            [
                os.path.abspath("setup.cfg"),
                os.path.abspath("tox.ini"),
                os.path.abspath(CLI_SPECIFIED_FILEPATH),
            ],
        ),
        # Missing extra config files specified
        (
            [
                CLI_SPECIFIED_FILEPATH,
                "tests/fixtures/config_files/missing.ini",
            ],
            [
                os.path.abspath("setup.cfg"),
                os.path.abspath("tox.ini"),
                os.path.abspath(CLI_SPECIFIED_FILEPATH),
            ],
        ),
    ],
)
def test_local_config_files(extra_config_files, expected):
    """Verify discovery of local config files."""
    finder = config.ConfigFileFinder("flake8", extra_config_files)

    assert list(finder.local_config_files()) == expected


def test_local_configs():
    """Verify we return a ConfigParser."""
    finder = config.ConfigFileFinder("flake8")

    assert isinstance(finder.local_configs(), configparser.RawConfigParser)


@pytest.mark.parametrize(
    "files",
    [
        [BROKEN_CONFIG_PATH],
        [CLI_SPECIFIED_FILEPATH, BROKEN_CONFIG_PATH],
    ],
)
def test_read_config_catches_broken_config_files(files):
    """Verify that we do not allow the exception to bubble up."""
    _, parsed = config.ConfigFileFinder._read_config(*files)
    assert BROKEN_CONFIG_PATH not in parsed


def test_read_config_catches_decoding_errors(tmpdir):
    """Verify that we do not allow the exception to bubble up."""
    setup_cfg = tmpdir.join("setup.cfg")
    # pick bytes that are unlikely to decode
    setup_cfg.write_binary(b"[x]\ny = \x81\x8d\x90\x9d")
    _, parsed = config.ConfigFileFinder._read_config(setup_cfg.strpath)
    assert parsed == []


def test_config_file_default_value():
    """Verify the default 'config_file' attribute value."""
    finder = config.ConfigFileFinder("flake8")
    assert finder.config_file is None


def test_setting_config_file_value():
    """Verify the 'config_file' attribute matches constructed value."""
    config_file_value = "flake8.ini"
    finder = config.ConfigFileFinder("flake8", config_file=config_file_value)
    assert finder.config_file == config_file_value


def test_ignore_config_files_default_value():
    """Verify the default 'ignore_config_files' attribute value."""
    finder = config.ConfigFileFinder("flake8")
    assert finder.ignore_config_files is False


@pytest.mark.parametrize(
    "ignore_config_files_arg",
    [
        False,
        True,
    ],
)
def test_setting_ignore_config_files_value(ignore_config_files_arg):
    """Verify the 'ignore_config_files' attribute matches constructed value."""
    finder = config.ConfigFileFinder(
        "flake8", ignore_config_files=ignore_config_files_arg
    )
    assert finder.ignore_config_files is ignore_config_files_arg



"""Unit tests for flake8.options.config.ConfigParser."""
import os
from unittest import mock

import pytest

from flake8.options import config
from flake8.options import manager


@pytest.fixture
def optmanager():
    """Generate an OptionManager with simple values."""
    return manager.OptionManager(prog="flake8", version="3.0.0a1")


@pytest.fixture
def config_finder():
    """Generate a simple ConfigFileFinder."""
    return config.ConfigFileFinder("flake8")


def test_parse_cli_config(optmanager, config_finder):
    """Parse the specified config file as a cli config file."""
    optmanager.add_option(
        "--exclude",
        parse_from_config=True,
        comma_separated_list=True,
        normalize_paths=True,
    )
    optmanager.add_option(
        "--ignore", parse_from_config=True, comma_separated_list=True
    )
    optmanager.add_option("--quiet", parse_from_config=True, action="count")
    parser = config.ConfigParser(optmanager, config_finder)

    config_file = "tests/fixtures/config_files/cli-specified.ini"
    parsed_config = parser.parse_cli_config(config_file)

    config_dir = os.path.dirname(config_file)
    assert parsed_config == {
        "ignore": ["E123", "W234", "E111"],
        "exclude": [
            os.path.abspath(os.path.join(config_dir, "foo/")),
            os.path.abspath(os.path.join(config_dir, "bar/")),
            os.path.abspath(os.path.join(config_dir, "bogus/")),
        ],
        "quiet": 1,
    }


@pytest.mark.parametrize(
    "filename,is_configured_by",
    [
        ("tests/fixtures/config_files/cli-specified.ini", True),
        ("tests/fixtures/config_files/no-flake8-section.ini", False),
    ],
)
def test_is_configured_by(
    filename, is_configured_by, optmanager, config_finder
):
    """Verify the behaviour of the is_configured_by method."""
    parsed_config, _ = config.ConfigFileFinder._read_config(filename)
    parser = config.ConfigParser(optmanager, config_finder)

    assert parser.is_configured_by(parsed_config) is is_configured_by


def test_parse_local_config(optmanager, config_finder):
    """Verify parsing of local config files."""
    optmanager.add_option(
        "--exclude",
        parse_from_config=True,
        comma_separated_list=True,
        normalize_paths=True,
    )
    optmanager.add_option(
        "--ignore", parse_from_config=True, comma_separated_list=True
    )
    optmanager.add_option("--quiet", parse_from_config=True, action="count")
    parser = config.ConfigParser(optmanager, config_finder)

    with mock.patch.object(config_finder, "local_config_files") as localcfs:
        localcfs.return_value = [
            "tests/fixtures/config_files/cli-specified.ini"
        ]
        parsed_config = parser.parse_local_config()

    assert parsed_config == {
        "ignore": ["E123", "W234", "E111"],
        "exclude": [
            os.path.abspath("foo/"),
            os.path.abspath("bar/"),
            os.path.abspath("bogus/"),
        ],
        "quiet": 1,
    }


def test_parse_isolates_config(optmanager):
    """Verify behaviour of the parse method with isolated=True."""
    config_finder = mock.MagicMock()
    config_finder.ignore_config_files = True
    parser = config.ConfigParser(optmanager, config_finder)

    assert parser.parse() == {}
    assert config_finder.local_configs.called is False


def test_parse_uses_cli_config(optmanager):
    """Verify behaviour of the parse method with a specified config."""
    config_file_value = "foo.ini"
    config_finder = mock.MagicMock()
    config_finder.config_file = config_file_value
    config_finder.ignore_config_files = False
    parser = config.ConfigParser(optmanager, config_finder)

    parser.parse()
    config_finder.cli_config.assert_called_once_with(config_file_value)


@pytest.mark.parametrize(
    "config_fixture_path",
    [
        "tests/fixtures/config_files/cli-specified.ini",
        "tests/fixtures/config_files/cli-specified-with-inline-comments.ini",
        "tests/fixtures/config_files/cli-specified-without-inline-comments.ini",  # noqa: E501
    ],
)
def test_parsed_configs_are_equivalent(
    optmanager, config_finder, config_fixture_path
):
    """Verify the each file matches the expected parsed output.

    This is used to ensure our documented behaviour does not regress.
    """
    optmanager.add_option(
        "--exclude",
        parse_from_config=True,
        comma_separated_list=True,
        normalize_paths=True,
    )
    optmanager.add_option(
        "--ignore", parse_from_config=True, comma_separated_list=True
    )
    parser = config.ConfigParser(optmanager, config_finder)

    with mock.patch.object(config_finder, "local_config_files") as localcfs:
        localcfs.return_value = [config_fixture_path]
        parsed_config = parser.parse()

    assert parsed_config["ignore"] == ["E123", "W234", "E111"]
    assert parsed_config["exclude"] == [
        os.path.abspath("foo/"),
        os.path.abspath("bar/"),
        os.path.abspath("bogus/"),
    ]


@pytest.mark.parametrize(
    "config_file",
    ["tests/fixtures/config_files/config-with-hyphenated-options.ini"],
)
def test_parsed_hyphenated_and_underscored_names(
    optmanager, config_finder, config_file
):
    """Verify we find hyphenated option names as well as underscored.

    This tests for options like --max-line-length and --enable-extensions
    which are able to be specified either as max-line-length or
    max_line_length in our config files.
    """
    optmanager.add_option(
        "--max-line-length", parse_from_config=True, type=int
    )
    optmanager.add_option(
        "--enable-extensions",
        parse_from_config=True,
        comma_separated_list=True,
    )
    parser = config.ConfigParser(optmanager, config_finder)

    with mock.patch.object(config_finder, "local_config_files") as localcfs:
        localcfs.return_value = [config_file]
        parsed_config = parser.parse()

    assert parsed_config["max_line_length"] == 110
    assert parsed_config["enable_extensions"] == ["H101", "H235"]



"""Tests for flake8.plugins.manager.PluginManager."""
from unittest import mock

from flake8._compat import importlib_metadata
from flake8.plugins import manager


@mock.patch.object(importlib_metadata, "entry_points")
def test_calls_entrypoints_on_instantiation(entry_points_mck):
    """Verify that we call entry_points() when we create a manager."""
    entry_points_mck.return_value = {}
    manager.PluginManager(namespace="testing.entrypoints")
    entry_points_mck.assert_called_once_with()


@mock.patch.object(importlib_metadata, "entry_points")
def test_calls_entrypoints_creates_plugins_automaticaly(entry_points_mck):
    """Verify that we create Plugins on instantiation."""
    entry_points_mck.return_value = {
        "testing.entrypoints": [
            importlib_metadata.EntryPoint("T100", "", "testing.entrypoints"),
            importlib_metadata.EntryPoint("T200", "", "testing.entrypoints"),
        ],
    }
    plugin_mgr = manager.PluginManager(namespace="testing.entrypoints")

    entry_points_mck.assert_called_once_with()
    assert "T100" in plugin_mgr.plugins
    assert "T200" in plugin_mgr.plugins
    assert isinstance(plugin_mgr.plugins["T100"], manager.Plugin)
    assert isinstance(plugin_mgr.plugins["T200"], manager.Plugin)


@mock.patch.object(importlib_metadata, "entry_points")
def test_handles_mapping_functions_across_plugins(entry_points_mck):
    """Verify we can use the PluginManager call functions on all plugins."""
    entry_points_mck.return_value = {
        "testing.entrypoints": [
            importlib_metadata.EntryPoint("T100", "", "testing.entrypoints"),
            importlib_metadata.EntryPoint("T200", "", "testing.entrypoints"),
        ],
    }
    plugin_mgr = manager.PluginManager(namespace="testing.entrypoints")
    plugins = [plugin_mgr.plugins[name] for name in plugin_mgr.names]

    assert list(plugin_mgr.map(lambda x: x)) == plugins


@mock.patch.object(importlib_metadata, "entry_points")
def test_local_plugins(entry_points_mck):
    """Verify PluginManager can load given local plugins."""
    entry_points_mck.return_value = {}
    plugin_mgr = manager.PluginManager(
        namespace="testing.entrypoints", local_plugins=["X = path.to:Plugin"]
    )

    assert plugin_mgr.plugins["X"].entry_point.value == "path.to:Plugin"



"""Tests for the Application class."""
import argparse
import sys
from unittest import mock

import pytest

from flake8.main import application as app


def options(**kwargs):
    """Generate argparse.Namespace for our Application."""
    kwargs.setdefault("verbose", 0)
    kwargs.setdefault("output_file", None)
    kwargs.setdefault("count", False)
    kwargs.setdefault("exit_zero", False)
    return argparse.Namespace(**kwargs)


@pytest.fixture
def application():
    """Create an application."""
    return app.Application()


@pytest.mark.parametrize(
    "result_count, catastrophic, exit_zero, value",
    [
        (0, False, False, False),
        (0, True, False, True),
        (2, False, False, True),
        (2, True, False, True),
        (0, True, True, True),
        (2, False, True, False),
        (2, True, True, True),
    ],
)
def test_exit_does_raise(
    result_count, catastrophic, exit_zero, value, application
):
    """Verify Application.exit doesn't raise SystemExit."""
    application.result_count = result_count
    application.catastrophic_failure = catastrophic
    application.options = options(exit_zero=exit_zero)

    with pytest.raises(SystemExit) as excinfo:
        application.exit()

    assert excinfo.value.args[0] is value


def test_warns_on_unknown_formatter_plugin_name(application):
    """Verify we log a warning with an unfound plugin."""
    default = mock.Mock()
    execute = default.execute
    application.formatting_plugins = {
        "default": default,
    }
    with mock.patch.object(app.LOG, "warning") as warning:
        assert execute is application.formatter_for("fake-plugin-name")

    assert warning.called is True
    assert warning.call_count == 1


def test_returns_specified_plugin(application):
    """Verify we get the plugin we want."""
    desired = mock.Mock()
    execute = desired.execute
    application.formatting_plugins = {
        "default": mock.Mock(),
        "desired": desired,
    }

    with mock.patch.object(app.LOG, "warning") as warning:
        assert execute is application.formatter_for("desired")

    assert warning.called is False


def test_prelim_opts_args(application):
    """Verify we get sensible prelim opts and args."""
    opts, args = application.parse_preliminary_options(
        ["--foo", "--verbose", "src", "setup.py", "--statistics", "--version"]
    )

    assert opts.verbose
    assert args == ["--foo", "src", "setup.py", "--statistics", "--version"]


def test_prelim_opts_ignore_help(application):
    """Verify -h/--help is not handled."""
    # GIVEN

    # WHEN
    _, args = application.parse_preliminary_options(["--help", "-h"])

    # THEN
    assert args == ["--help", "-h"]


def test_prelim_opts_handles_empty(application):
    """Verify empty argv lists are handled correctly."""
    irrelevant_args = ["myexe", "/path/to/foo"]
    with mock.patch.object(sys, "argv", irrelevant_args):
        opts, args = application.parse_preliminary_options([])

        assert args == []



"""Unit tests for the FileChecker class."""
from unittest import mock

import pytest

import flake8
from flake8 import checker


@mock.patch("flake8.checker.FileChecker._make_processor", return_value=None)
def test_repr(*args):
    """Verify we generate a correct repr."""
    file_checker = checker.FileChecker(
        "example.py",
        checks={},
        options=object(),
    )
    assert repr(file_checker) == "FileChecker for example.py"


def test_nonexistent_file():
    """Verify that checking non-existent file results in an error."""
    c = checker.FileChecker("foobar.py", checks={}, options=object())

    assert c.processor is None
    assert not c.should_process
    assert len(c.results) == 1
    error = c.results[0]
    assert error[0] == "E902"


def test_raises_exception_on_failed_plugin(tmp_path, default_options):
    """Checks that a failing plugin results in PluginExecutionFailed."""
    foobar = tmp_path / "foobar.py"
    foobar.write_text("I exist!")  # Create temp file
    plugin = {
        "name": "failure",
        "plugin_name": "failure",  # Both are necessary
        "parameters": dict(),
        "plugin": mock.MagicMock(side_effect=ValueError),
    }
    """Verify a failing plugin results in an plugin error"""
    fchecker = checker.FileChecker(
        str(foobar), checks=[], options=default_options
    )
    with pytest.raises(flake8.exceptions.PluginExecutionFailed):
        fchecker.run_check(plugin)



"""Tests for the flake8.style_guide.DecisionEngine class."""
import argparse

import pytest

from flake8 import defaults
from flake8 import style_guide


def create_options(**kwargs):
    """Create and return an instance of argparse.Namespace."""
    kwargs.setdefault("select", [])
    kwargs.setdefault("extended_default_ignore", [])
    kwargs.setdefault("extended_default_select", [])
    kwargs.setdefault("extend_select", [])
    kwargs.setdefault("ignore", [])
    kwargs.setdefault("extend_ignore", [])
    kwargs.setdefault("disable_noqa", False)
    kwargs.setdefault("enable_extensions", [])
    return argparse.Namespace(**kwargs)


@pytest.mark.parametrize(
    "ignore_list,extend_ignore,error_code",
    [
        (["E111", "E121"], [], "E111"),
        (["E111", "E121"], [], "E121"),
        (["E111"], ["E121"], "E121"),
        (["E11", "E12"], [], "E121"),
        (["E2", "E12"], [], "E121"),
        (["E2", "E12"], [], "E211"),
        (["E2", "E3"], ["E12"], "E211"),
    ],
)
def test_was_ignored_ignores_errors(ignore_list, extend_ignore, error_code):
    """Verify we detect users explicitly ignoring an error."""
    decider = style_guide.DecisionEngine(
        create_options(ignore=ignore_list, extend_ignore=extend_ignore)
    )

    assert decider.was_ignored(error_code) is style_guide.Ignored.Explicitly


@pytest.mark.parametrize(
    "ignore_list,extend_ignore,error_code",
    [
        (["E111", "E121"], [], "E112"),
        (["E111", "E121"], [], "E122"),
        (["E11", "E12"], ["E121"], "W121"),
        (["E2", "E12"], [], "E112"),
        (["E2", "E12"], [], "E111"),
        (["E2", "E12"], ["W11", "E3"], "E111"),
    ],
)
def test_was_ignored_implicitly_selects_errors(
    ignore_list, extend_ignore, error_code
):
    """Verify we detect users does not explicitly ignore an error."""
    decider = style_guide.DecisionEngine(
        create_options(ignore=ignore_list, extend_ignore=extend_ignore)
    )

    assert decider.was_ignored(error_code) is style_guide.Selected.Implicitly


@pytest.mark.parametrize(
    "select_list,extend_select,enable_extensions,error_code",
    [
        (["E111", "E121"], [], [], "E111"),
        (["E111", "E121"], [], [], "E121"),
        (["E11", "E12"], [], [], "E121"),
        (["E2", "E12"], [], [], "E121"),
        (["E2", "E12"], [], [], "E211"),
        (["E1"], ["E2"], [], "E211"),
        (["E1"], [], ["E2"], "E211"),
        ([], ["E2"], [], "E211"),
        ([], [], ["E2"], "E211"),
        (["E1"], ["E2"], [], "E211"),
        (["E111"], ["E121"], ["E2"], "E121"),
    ],
)
def test_was_selected_selects_errors(
    select_list, extend_select, enable_extensions, error_code
):
    """Verify we detect users explicitly selecting an error."""
    decider = style_guide.DecisionEngine(
        options=create_options(
            select=select_list,
            extend_select=extend_select,
            enable_extensions=enable_extensions,
        ),
    )

    assert decider.was_selected(error_code) is style_guide.Selected.Explicitly


def test_was_selected_implicitly_selects_errors():
    """Verify we detect users implicitly selecting an error."""
    error_code = "E121"
    decider = style_guide.DecisionEngine(
        create_options(
            select=[],
            extended_default_select=["E"],
        ),
    )

    assert decider.was_selected(error_code) is style_guide.Selected.Implicitly


@pytest.mark.parametrize(
    "select_list,error_code",
    [
        (["E111", "E121"], "E112"),
        (["E111", "E121"], "E122"),
        (["E11", "E12"], "E132"),
        (["E2", "E12"], "E321"),
        (["E2", "E12"], "E410"),
    ],
)
def test_was_selected_excludes_errors(select_list, error_code):
    """Verify we detect users implicitly excludes an error."""
    decider = style_guide.DecisionEngine(create_options(select=select_list))

    assert decider.was_selected(error_code) is style_guide.Ignored.Implicitly


@pytest.mark.parametrize(
    "select_list,ignore_list,extend_ignore,error_code,expected",
    [
        (["E111", "E121"], [], [], "E111", style_guide.Decision.Selected),
        (["E111", "E121"], [], [], "E112", style_guide.Decision.Ignored),
        (["E111", "E121"], [], [], "E121", style_guide.Decision.Selected),
        (["E111", "E121"], [], [], "E122", style_guide.Decision.Ignored),
        (["E11", "E12"], [], [], "E132", style_guide.Decision.Ignored),
        (["E2", "E12"], [], [], "E321", style_guide.Decision.Ignored),
        (["E2", "E12"], [], [], "E410", style_guide.Decision.Ignored),
        (["E11", "E121"], ["E1"], [], "E112", style_guide.Decision.Selected),
        (["E11", "E121"], [], ["E1"], "E112", style_guide.Decision.Selected),
        (
            ["E111", "E121"],
            ["E2"],
            ["E3"],
            "E122",
            style_guide.Decision.Ignored,
        ),
        (["E11", "E12"], ["E13"], [], "E132", style_guide.Decision.Ignored),
        (["E1", "E3"], ["E32"], [], "E321", style_guide.Decision.Ignored),
        ([], ["E2", "E12"], [], "E410", style_guide.Decision.Ignored),
        (
            ["E4"],
            ["E2", "E12", "E41"],
            [],
            "E410",
            style_guide.Decision.Ignored,
        ),
        (
            ["E41"],
            ["E2", "E12", "E4"],
            [],
            "E410",
            style_guide.Decision.Selected,
        ),
        (["E"], ["F"], [], "E410", style_guide.Decision.Selected),
        (["F"], [], [], "E410", style_guide.Decision.Ignored),
        (["E"], defaults.IGNORE, [], "E126", style_guide.Decision.Selected),
        (["W"], defaults.IGNORE, [], "E126", style_guide.Decision.Ignored),
        (["E"], defaults.IGNORE, [], "W391", style_guide.Decision.Ignored),
        (["E", "W"], ["E13"], [], "E131", style_guide.Decision.Ignored),
        (defaults.SELECT, ["E13"], [], "E131", style_guide.Decision.Ignored),
        (
            defaults.SELECT,
            defaults.IGNORE,
            ["W391"],
            "E126",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            defaults.IGNORE,
            [],
            "W391",
            style_guide.Decision.Selected,
        ),
    ],
)
def test_decision_for(
    select_list, ignore_list, extend_ignore, error_code, expected
):
    """Verify we decide when to report an error."""
    decider = style_guide.DecisionEngine(
        create_options(
            select=select_list,
            ignore=ignore_list,
            extend_ignore=extend_ignore,
        )
    )

    assert decider.decision_for(error_code) is expected


@pytest.mark.parametrize(
    "select,ignore,extended_default_ignore,extended_default_select,"
    "enabled_extensions,error_code,expected",
    [
        (
            defaults.SELECT,
            [],
            [],
            ["I1"],
            [],
            "I100",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            [],
            [],
            ["I1"],
            [],
            "I201",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            ["I2"],
            [],
            ["I1"],
            [],
            "I101",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            ["I2"],
            [],
            ["I1"],
            [],
            "I201",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            ["I1"],
            [],
            ["I10"],
            [],
            "I101",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            ["I10"],
            [],
            ["I1"],
            [],
            "I101",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            [],
            [],
            [],
            ["U4"],
            "U401",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            ["U401"],
            [],
            [],
            ["U4"],
            "U401",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            ["U401"],
            [],
            [],
            ["U4"],
            "U402",
            style_guide.Decision.Selected,
        ),
        (
            ["E", "W"],
            ["E13"],
            [],
            [],
            [],
            "E131",
            style_guide.Decision.Ignored,
        ),
        (
            ["E", "W"],
            ["E13"],
            [],
            [],
            [],
            "E126",
            style_guide.Decision.Selected,
        ),
        (["E2"], ["E21"], [], [], [], "E221", style_guide.Decision.Selected),
        (["E2"], ["E21"], [], [], [], "E212", style_guide.Decision.Ignored),
        (
            ["F", "W"],
            ["C90"],
            [],
            ["I1"],
            [],
            "C901",
            style_guide.Decision.Ignored,
        ),
        (
            ["E", "W"],
            ["C"],
            [],
            [],
            [],
            "E131",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            defaults.IGNORE,
            [],
            [],
            ["I"],
            "I101",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            defaults.IGNORE,
            [],
            ["G"],
            ["I"],
            "G101",
            style_guide.Decision.Selected,
        ),
        (
            defaults.SELECT,
            ["G1"],
            [],
            ["G"],
            ["I"],
            "G101",
            style_guide.Decision.Ignored,
        ),
        (
            defaults.SELECT,
            ["E126"],
            [],
            [],
            ["I"],
            "I101",
            style_guide.Decision.Selected,
        ),
        (
            ["E", "W"],
            defaults.IGNORE,
            [],
            ["I"],
            [],
            "I101",
            style_guide.Decision.Ignored,
        ),
        (
            ["E", "W", "I101"],
            defaults.IGNORE + ("I101",),
            ["I101"],
            [],
            [],
            "I101",
            style_guide.Decision.Selected,
        ),
        (
            ["E", "W"],
            defaults.IGNORE + ("I101",),
            ["I101"],
            [],
            [],
            "I101",
            style_guide.Decision.Ignored,
        ),
        # TODO(sigmavirus24) Figure out how to exercise the final catch-all
        # return statement
    ],
)
def test_more_specific_decision_for_logic(
    select,
    ignore,
    extended_default_ignore,
    extended_default_select,
    enabled_extensions,
    error_code,
    expected,
):
    """Verify the logic of DecisionEngine.more_specific_decision_for."""
    decider = style_guide.DecisionEngine(
        create_options(
            select=select,
            ignore=ignore,
            extended_default_select=extended_default_select,
            extended_default_ignore=extended_default_ignore,
            enable_extensions=enabled_extensions,
        ),
    )

    assert decider.more_specific_decision_for(error_code) is expected
